#!/usr/bin/env python3

import asyncio
import json
import logging
import pyautogui

logging.basicConfig(level=logging.INFO)

PERIOD = 0.1

# For now, messages will be in json format. Each message will contain an array
# of commands, and each command has a different format.

# Commands:
#   set_distance: <float> -- Set the goal distance
#   distance: <float>     -- Report the current target distance
#   enable: (1 or 0)
#   set_p: <float>        -- Sets the "proportional" gain value
#   set_d: <float>        -- Sets the "derivative" gain value
#   set_threshold: float  -- sets control signal threshold

# i.e. A command to set the distance and enable would be the json string:
#    "[ {set_distance: 232.2}, {enable: 1} ]"

# *sigh* Gotta use Python 3.6 because Ubuntu+pyautogui
class EcaProtocol(asyncio.Protocol):
    def connection_made(self, transport):
        logging.info('Received connection.')
        self.transport = transport

        self.set_distance = 250
        self.p = 10
        self.d = 10
        self.key_down = None
        self.forward_key = 'w'
        self.backward_key = 's'
        self.enabled = False
        self.last_distance = 0
        self.threshold = 50

    def connection_lost(self, exc):
        logging.info('Lost connection.')
        self.enabled = False

    def _handle_enable(self, enable):
        if enable:
            logging.info('Enabled.')
            self.enabled = True
        else:
            logging.info('Disabled.')
            if self.key_down:
                pyautogui.keyUp(self.key_down)
                self.key_down = None

    def _handle_set_distance(self, distance):
        logging.info('Set distance: {}'.format(distance))
        self.set_distance = float(distance)

    def _handle_set_p(self, p):
        logging.info('Set p: {}'.format(p))
        self.p = float(p)

    def _handle_set_d(self, d):
        logging.info('set d: {}'.format(d))
        self.d = float(d)

    def _handle_set_threshold(self, thresh):
        logging.info('set threshold: {}'.format(thresh))
        self.threshold = float(thresh)

    def __accel(self):
        if self.key_down is not self.forward_key:
            if self.key_down is not None:
                pyautogui.keyUp(self.key_down)
            pyautogui.keyDown(self.forward_key)
            self.key_down = self.forward_key

    def __decel(self):
        if self.key_down is not self.backward_key:
            if self.key_down is not None:
                pyautogui.keyUp(self.key_down)
            pyautogui.keyDown(self.backward_key)
            self.key_down = self.backward_key

    def __neutral(self):
        if self.key_down:
            pyautogui.keyUp(self.key_down)
            self.key_down = None

    def _handle_distance(self, distance):
        # Received a distance update from Lua. Run PID controller
        v = (distance - self.last_distance) / PERIOD
        self.last_distance = distance

        e = distance - self.set_distance

        # Sig should be positive if we need to get closer, negative otherwise
        sig = e*self.p + v*self.d

        logging.debug('distance: {}, v: {}, sig: {}'.format(
            distance, v, sig) )

        if sig > self.threshold:
            self.__accel()
        elif sig < -self.threshold:
            self.__decel()
        else:
            self.__neutral()

    def data_received(self, data):
        logging.debug('Received data! {}'.format(data))
        try:
            obj = json.loads(data)
        except Exception as e:
            logging.error('Could not decode json: {}:{}'.format(data, e))
            return

        handlers = {
            'set_distance': self._handle_set_distance,
            'distance': self._handle_distance,
            'enable': self._handle_enable,
            'set_p': self._handle_set_p,
            'set_d': self._handle_set_d,
            'set_threshold': self._handle_set_threshold
        }

        for item in obj:
            for k,v in item.items():
                try:
                    handlers[k](v)
                except Exception as e:
                    logging.error('Could not process command: {}: {}'.format(k, e))


if __name__ == '__main__':
    # Start the server
    loop = asyncio.get_event_loop()
    coro = loop.create_server(EcaProtocol, 'localhost', 9090)
    logging.info('Starting server...')
    server = loop.run_until_complete(coro)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close
    
